/**
 *  For each HTML file with an AST syntax highlight code blocks that have a
 *  class name indicating the source code language.
 *
 *  @class Highlight
 *
 *  @see https://highlightjs.org/ Highlight.js
 *  @see http://highlightjs.readthedocs.io/en/latest/api.html Highlight.js API
 */
class Highlight {

  /**
   *  Create a Highlight plugin.
   *
   *  The default `selector` when none is given is `code[class]`.
   *
   *  The options specific to `highlight.js` are equivalent to the default
   *  values except that `classPrefix` is set to the empty string so class names
   *  will not contain the `hljs-` prefix.
   *
   *  @constructor Highlight
   *  @param {Object} context the processing context.
   *  @param {Object} options plugin options.
   *
   *  @option {String} [selector] query for code blocks.
   *  @option {Function} [language] extract language from class names.
   *  @option {Function} [highlighter] override the default highlighter.
   *  @option {String} [classPrefix] highlight.js `classPrefix` option.
   *  @option {String} [tabReplace] highlight.js `tabReplace` option.
   *  @option {Boolean} [useBR] highlight.js `useBR` option.
   *  @option {Array} [languages] highlight.js `languages` option.
   *
   */
  constructor (context, options = {}) {
    this.context = context
    this.options = options
    this.selector = options.selector || 'code[class]'
    this.language = typeof options.language === 'function'
      ? options.language : this.getLanguage
    this.highlighter = typeof options.highlighter === 'function'
      ? options.highlighter : this.highlight
    this.classPrefix = options.classPrefix !== undefined
      ? options.classPrefix : ''
    this.tabReplace = options.tabReplace
    this.useBR = options.useBR
    this.languages = options.languages
  }

  /**
   *  Default implementation to extract a language identifier for an element.
   *
   *  Use the `language` option to supply an alternative implementation.
   *
   *  Finds the first `lang-` or `language-` class name on the element and
   *  returns the language identifier.
   *
   *  @function getLanguage
   *  @member Highlight
   *  @param {Object} file the current file.
   *  @param {Object} element the matched element.
   *
   *  @returns string language identifier.
   */
  getLanguage (file, element) {
    const ast = file.ast.html
    const classNames = ast.adapter.getAttributeValue(element, 'class')
    const classes = classNames.split(/\s+/)
    const ptn = /^lang(uage)?-/
    let lang, name
    for (let i = 0; i < classes.length; i++) {
      name = classes[i]
      if (ptn.test(name)) {
        lang = name.replace(ptn, '')
        break
      }
    }
    return lang
  }

  /**
   *  Default implementation to highlight code blocks.
   *
   *  Use the `highlighter` option to supply an alternative implementation.
   *
   *  @function highlight
   *  @member Highlight
   *  @param {Object} file the current file.
   *  @param {Object} element the matched element.
   *  @param {String} lang the language identifier.
   *  @param {String} text content of the code block.
   *
   *  @returns string highlighted code.
   */
  highlight (file, element, lang, text) {
    const highlighter = require('highlightjs')
    highlighter.configure(this)
    const result = highlighter.highlight(lang, text)
    return result.value
  }

  /**
   *  For each file with an HTML AST find elements that match the `selector`
   *  option. Call the `language` function for each matched element and if a
   *  language string is returned call the `highlighter` function.
   *
   *  Use the return value from the `highlighter` function as a DOM document
   *  fragment and update the code block with the new child nodes.
   *
   *  @function sources
   *  @member Highlight
   *  @param {File} file the current file.
   *  @param {Object} context the processing context.
   *  @param {Object} options plugin options.
   */
  sources (file, context, options = {}) {
    const log = context.log
    const ast = file.ast.html

    if (!ast) {
      return log.warn('[highlight] missing HTML document AST for %s', file.path)
    }

    const elements = ast.querySelectorAll(this.selector)
    elements.forEach((el) => {
      const lang = this.language(file, el)
      if (lang) {
        let text = ast.adapter.getText(el)
        let result = this.highlighter(file, el, lang, text)
        let fragment = ast.parseFragment(result)
        if (fragment.childNodes.length) {
          // remove existing children
          el.childNodes = []
          // append fragment nodes
          fragment.childNodes.forEach((node) => {
            ast.adapter.appendChild(el, node)
          })
          ast.dirty = true
        }
      }
    })
  }

  static get exclude () {
    return [/google[a-z0-9]+\.html$/i]
  }

  static get test () {
    return /\.(sgr|jade|ejs|html)$/
  }
}

module.exports = Highlight
